@@ -2,8 +2,8 @@ use std::ops::{Deref, DerefMut};
 
 use serde::de::value::{MapDeserializer, SeqDeserializer};
 use serde::de::{
-    self, Deserialize, DeserializeSeed, Deserializer, EnumAccess, IntoDeserializer, MapAccess,
-    SeqAccess, Unexpected, VariantAccess, Visitor,
+    self, Deserialize, DeserializeOwned, DeserializeSeed, Deserializer, EnumAccess,
+    IntoDeserializer, MapAccess, SeqAccess, Unexpected, VariantAccess, Visitor,
 };
 use serde::forward_to_deserialize_any;
 
@@ -110,9 +110,9 @@ impl<'de> Visitor<'de> for ValueVisitor {
 /// env.add_filter("dirname", dirname);
 /// ```
 #[cfg_attr(docsrs, doc(cfg(feature = "deserialization")))]
-pub struct ViaDeserialize<T>(pub T);
+pub struct ViaDeserialize<T: DeserializeOwned>(pub T);
 
-impl<'a, T: Deserialize<'a>> ArgType<'a> for ViaDeserialize<T> {
+impl<'a, T: DeserializeOwned> ArgType<'a> for ViaDeserialize<T> {
     type Output = Self;
 
     fn from_value(value: Option<&'a Value>) -> Result<Self, Error> {
@@ -131,7 +131,7 @@ impl<'a, T: Deserialize<'a>> ArgType<'a> for ViaDeserialize<T> {
     }
 }
 
-impl<T> Deref for ViaDeserialize<T> {
+impl<T: DeserializeOwned> Deref for ViaDeserialize<T> {
     type Target = T;
 
     fn deref(&self) -> &T {
@@ -139,7 +139,7 @@ impl<T> Deref for ViaDeserialize<T> {
     }
 }
 
-impl<T> DerefMut for ViaDeserialize<T> {
+impl<T: DeserializeOwned> DerefMut for ViaDeserialize<T> {
     fn deref_mut(&mut self) -> &mut T {
         &mut self.0
     }
@@ -158,7 +158,7 @@ macro_rules! common_forward {
 }
 
 #[cfg_attr(docsrs, doc(cfg(feature = "deserialization")))]
-impl<'de> IntoDeserializer<'de, Error> for Value {
+impl IntoDeserializer<'_, Error> for Value {
     type Deserializer = Value;
 
     fn into_deserializer(self) -> Value {
@@ -181,7 +181,7 @@ impl<'de> Deserializer<'de> for Value {
             ValueRepr::F64(v) => visitor.visit_f64(v),
             ValueRepr::String(ref v, _) => visitor.visit_str(v),
             ValueRepr::SmallStr(v) => visitor.visit_str(v.as_str()),
-            ValueRepr::Undefined | ValueRepr::None => visitor.visit_unit(),
+            ValueRepr::Undefined(_) | ValueRepr::None => visitor.visit_unit(),
             ValueRepr::Bytes(ref v) => visitor.visit_bytes(v),
             ValueRepr::Object(o) => match o.repr() {
                 ObjectRepr::Plain => Err(de::Error::custom("cannot deserialize plain objects")),
@@ -197,7 +197,7 @@ impl<'de> Deserializer<'de> for Value {
 
     fn deserialize_option<V: Visitor<'de>>(self, visitor: V) -> Result<V::Value, Error> {
         match self.0 {
-            ValueRepr::None | ValueRepr::Undefined => visitor.visit_unit(),
+            ValueRepr::None | ValueRepr::Undefined(_) => visitor.visit_unit(),
             _ => visitor.visit_some(self),
         }
     }
@@ -343,7 +343,7 @@ impl<'de> VariantAccess<'de> for VariantDeserializer {
 }
 
 #[cfg_attr(docsrs, doc(cfg(feature = "deserialization")))]
-impl<'de, 'v> Deserializer<'de> for &'v Value {
+impl<'de> Deserializer<'de> for &Value {
     type Error = Error;
 
     #[inline]
@@ -403,9 +403,9 @@ impl de::Error for Error {
     }
 }
 
-fn value_to_unexpected(value: &Value) -> Unexpected {
+fn value_to_unexpected(value: &Value) -> Unexpected<'_> {
     match value.0 {
-        ValueRepr::Undefined | ValueRepr::None => Unexpected::Unit,
+        ValueRepr::Undefined(_) | ValueRepr::None => Unexpected::Unit,
         ValueRepr::Bool(val) => Unexpected::Bool(val),
         ValueRepr::U64(val) => Unexpected::Unsigned(val),
         ValueRepr::I64(val) => Unexpected::Signed(val),
