@@ -3,14 +3,15 @@ use std::fmt;
 
 use serde::{ser, Serialize, Serializer};
 
+use crate::error::{Error, ErrorKind};
 use crate::utils::untrusted_size_hint;
 use crate::value::{
-    value_map_with_capacity, Arc, KeyRef, MapType, Packed, StringType, Value, ValueMap, ValueRepr,
-    VALUE_HANDLES, VALUE_HANDLE_MARKER,
+    value_map_with_capacity, Arc, Packed, Value, ValueMap, ValueRepr, VALUE_HANDLES,
+    VALUE_HANDLE_MARKER,
 };
 
 #[derive(Debug)]
-pub struct InvalidValue(Arc<str>);
+pub struct InvalidValue(String);
 
 impl std::error::Error for InvalidValue {}
 
@@ -25,7 +26,7 @@ impl serde::ser::Error for InvalidValue {
     where
         T: fmt::Display,
     {
-        InvalidValue(Arc::from(msg.to_string()))
+        InvalidValue(msg.to_string())
     }
 }
 
@@ -36,7 +37,7 @@ impl serde::ser::Error for InvalidValue {
 pub fn transform<T: Serialize>(value: T) -> Value {
     match value.serialize(ValueSerializer) {
         Ok(rv) => rv,
-        Err(invalid) => ValueRepr::Invalid(invalid.0).into(),
+        Err(invalid) => Value::from(Error::new(ErrorKind::BadSerialization, invalid.0)),
     }
 }
 
@@ -107,11 +108,11 @@ impl Serializer for ValueSerializer {
     }
 
     fn serialize_char(self, v: char) -> Result<Value, InvalidValue> {
-        Ok(ValueRepr::String(Arc::from(v.to_string()), StringType::Normal).into())
+        Ok(Value::from(v))
     }
 
     fn serialize_str(self, value: &str) -> Result<Value, InvalidValue> {
-        Ok(ValueRepr::String(Arc::from(value.to_owned()), StringType::Normal).into())
+        Ok(Value::from(value))
     }
 
     fn serialize_bytes(self, value: &[u8]) -> Result<Value, InvalidValue> {
@@ -122,9 +123,9 @@ impl Serializer for ValueSerializer {
         Ok(ValueRepr::None.into())
     }
 
-    fn serialize_some<T: ?Sized>(self, value: &T) -> Result<Value, InvalidValue>
+    fn serialize_some<T>(self, value: &T) -> Result<Value, InvalidValue>
     where
-        T: Serialize,
+        T: Serialize + ?Sized,
     {
         Ok(transform(value))
     }
@@ -139,34 +140,25 @@ impl Serializer for ValueSerializer {
 
     fn serialize_unit_variant(
         self,
-        name: &'static str,
-        variant_index: u32,
+        _name: &'static str,
+        _variant_index: u32,
         variant: &'static str,
     ) -> Result<Value, InvalidValue> {
-        if name == VALUE_HANDLE_MARKER && variant == VALUE_HANDLE_MARKER {
-            Ok(VALUE_HANDLES.with(|handles| {
-                let mut handles = handles.borrow_mut();
-                handles
-                    .remove(&variant_index)
-                    .expect("value handle not in registry")
-            }))
-        } else {
-            Ok(Value::from(variant))
-        }
+        Ok(Value::from(variant))
     }
 
-    fn serialize_newtype_struct<T: ?Sized>(
+    fn serialize_newtype_struct<T>(
         self,
         _name: &'static str,
         value: &T,
     ) -> Result<Value, InvalidValue>
     where
-        T: Serialize,
+        T: Serialize + ?Sized,
     {
         Ok(transform(value))
     }
 
-    fn serialize_newtype_variant<T: ?Sized>(
+    fn serialize_newtype_variant<T>(
         self,
         _name: &'static str,
         _variant_index: u32,
@@ -174,11 +166,11 @@ impl Serializer for ValueSerializer {
         value: &T,
     ) -> Result<Value, InvalidValue>
     where
-        T: Serialize,
+        T: Serialize + ?Sized,
     {
         let mut map = value_map_with_capacity(1);
-        map.insert(KeyRef::Str(variant), transform(value));
-        Ok(ValueRepr::Map(Arc::new(map), MapType::Normal).into())
+        map.insert(variant.into(), transform(value));
+        Ok(Value::from_object(map))
     }
 
     fn serialize_seq(self, len: Option<usize>) -> Result<Self::SerializeSeq, InvalidValue> {
@@ -195,11 +187,13 @@ impl Serializer for ValueSerializer {
 
     fn serialize_tuple_struct(
         self,
-        _name: &'static str,
+        name: &'static str,
         len: usize,
     ) -> Result<Self::SerializeTupleStruct, InvalidValue> {
-        Ok(SerializeTupleStruct {
-            fields: Vec::with_capacity(untrusted_size_hint(len)),
+        Ok(if name == VALUE_HANDLE_MARKER {
+            SerializeTupleStruct::Handle(None)
+        } else {
+            SerializeTupleStruct::Fields(Vec::with_capacity(untrusted_size_hint(len)))
         })
     }
 
@@ -255,16 +249,16 @@ impl ser::SerializeSeq for SerializeSeq {
     type Ok = Value;
     type Error = InvalidValue;
 
-    fn serialize_element<T: ?Sized>(&mut self, value: &T) -> Result<(), InvalidValue>
+    fn serialize_element<T>(&mut self, value: &T) -> Result<(), InvalidValue>
     where
-        T: Serialize,
+        T: Serialize + ?Sized,
     {
         self.elements.push(transform(value));
         Ok(())
     }
 
     fn end(self) -> Result<Value, InvalidValue> {
-        Ok(ValueRepr::Seq(Arc::new(self.elements)).into())
+        Ok(Value::from_object(self.elements))
     }
 }
 
@@ -276,37 +270,52 @@ impl ser::SerializeTuple for SerializeTuple {
     type Ok = Value;
     type Error = InvalidValue;
 
-    fn serialize_element<T: ?Sized>(&mut self, value: &T) -> Result<(), InvalidValue>
+    fn serialize_element<T>(&mut self, value: &T) -> Result<(), InvalidValue>
     where
-        T: Serialize,
+        T: Serialize + ?Sized,
     {
         self.elements.push(transform(value));
         Ok(())
     }
 
     fn end(self) -> Result<Value, InvalidValue> {
-        Ok(ValueRepr::Seq(Arc::new(self.elements)).into())
+        Ok(Value::from_object(self.elements))
     }
 }
 
-pub struct SerializeTupleStruct {
-    fields: Vec<Value>,
+pub enum SerializeTupleStruct {
+    Handle(Option<u32>),
+    Fields(Vec<Value>),
 }
 
 impl ser::SerializeTupleStruct for SerializeTupleStruct {
     type Ok = Value;
     type Error = InvalidValue;
 
-    fn serialize_field<T: ?Sized>(&mut self, value: &T) -> Result<(), InvalidValue>
+    fn serialize_field<T>(&mut self, value: &T) -> Result<(), InvalidValue>
     where
-        T: Serialize,
+        T: Serialize + ?Sized,
     {
-        self.fields.push(transform(value));
+        match self {
+            SerializeTupleStruct::Handle(ref mut handle) => {
+                *handle = transform(value).as_usize().map(|x| x as u32);
+            }
+            SerializeTupleStruct::Fields(ref mut fields) => {
+                fields.push(transform(value));
+            }
+        }
         Ok(())
     }
 
     fn end(self) -> Result<Value, InvalidValue> {
-        Ok(Value(ValueRepr::Seq(Arc::new(self.fields))))
+        match self {
+            SerializeTupleStruct::Handle(handle) => VALUE_HANDLES.with(|handles| {
+                handle
+                    .and_then(|h| handles.borrow_mut().remove(&h))
+                    .ok_or_else(|| InvalidValue("value handle not in registry".into()))
+            }),
+            SerializeTupleStruct::Fields(fields) => Ok(Value::from_object(fields)),
+        }
     }
 }
 
@@ -319,9 +328,9 @@ impl ser::SerializeTupleVariant for SerializeTupleVariant {
     type Ok = Value;
     type Error = InvalidValue;
 
-    fn serialize_field<T: ?Sized>(&mut self, value: &T) -> Result<(), InvalidValue>
+    fn serialize_field<T>(&mut self, value: &T) -> Result<(), InvalidValue>
     where
-        T: Serialize,
+        T: Serialize + ?Sized,
     {
         self.fields.push(transform(value));
         Ok(())
@@ -329,11 +338,8 @@ impl ser::SerializeTupleVariant for SerializeTupleVariant {
 
     fn end(self) -> Result<Value, InvalidValue> {
         let mut map = value_map_with_capacity(1);
-        map.insert(
-            KeyRef::Str(self.name),
-            Value(ValueRepr::Seq(self.fields.into())),
-        );
-        Ok(Value(ValueRepr::Map(map.into(), MapType::Normal)))
+        map.insert(self.name.into(), Value::from_object(self.fields));
+        Ok(Value::from_object(map))
     }
 }
 
@@ -346,9 +352,9 @@ impl ser::SerializeMap for SerializeMap {
     type Ok = Value;
     type Error = InvalidValue;
 
-    fn serialize_key<T: ?Sized>(&mut self, key: &T) -> Result<(), InvalidValue>
+    fn serialize_key<T>(&mut self, key: &T) -> Result<(), InvalidValue>
     where
-        T: Serialize,
+        T: Serialize + ?Sized,
     {
         match key.serialize(ValueSerializer) {
             Ok(key) => self.key = Some(key),
@@ -357,34 +363,27 @@ impl ser::SerializeMap for SerializeMap {
         Ok(())
     }
 
-    fn serialize_value<T: ?Sized>(&mut self, value: &T) -> Result<(), InvalidValue>
+    fn serialize_value<T>(&mut self, value: &T) -> Result<(), InvalidValue>
     where
-        T: Serialize,
+        T: Serialize + ?Sized,
     {
         if let Some(key) = self.key.take() {
-            self.entries.insert(KeyRef::Value(key), transform(value));
+            self.entries.insert(key, transform(value));
         }
         Ok(())
     }
 
     fn end(self) -> Result<Value, InvalidValue> {
-        Ok(Value(ValueRepr::Map(
-            Arc::new(self.entries),
-            MapType::Normal,
-        )))
+        Ok(Value::from_object(self.entries))
     }
 
-    fn serialize_entry<K: ?Sized, V: ?Sized>(
-        &mut self,
-        key: &K,
-        value: &V,
-    ) -> Result<(), InvalidValue>
+    fn serialize_entry<K, V>(&mut self, key: &K, value: &V) -> Result<(), InvalidValue>
     where
-        K: Serialize,
-        V: Serialize,
+        K: Serialize + ?Sized,
+        V: Serialize + ?Sized,
     {
         if let Ok(key) = key.serialize(ValueSerializer) {
-            self.entries.insert(KeyRef::Value(key), transform(value));
+            self.entries.insert(key, transform(value));
         }
         Ok(())
     }
@@ -398,20 +397,16 @@ impl ser::SerializeStruct for SerializeStruct {
     type Ok = Value;
     type Error = InvalidValue;
 
-    fn serialize_field<T: ?Sized>(
-        &mut self,
-        key: &'static str,
-        value: &T,
-    ) -> Result<(), InvalidValue>
+    fn serialize_field<T>(&mut self, key: &'static str, value: &T) -> Result<(), InvalidValue>
     where
-        T: Serialize,
+        T: Serialize + ?Sized,
     {
-        self.fields.insert(KeyRef::Str(key), transform(value));
+        self.fields.insert(key.into(), transform(value));
         Ok(())
     }
 
     fn end(self) -> Result<Value, InvalidValue> {
-        Ok(ValueRepr::Map(Arc::new(self.fields), MapType::Normal).into())
+        Ok(Value::from_object(self.fields))
     }
 }
 
@@ -424,24 +419,17 @@ impl ser::SerializeStructVariant for SerializeStructVariant {
     type Ok = Value;
     type Error = InvalidValue;
 
-    fn serialize_field<T: ?Sized>(
-        &mut self,
-        key: &'static str,
-        value: &T,
-    ) -> Result<(), InvalidValue>
+    fn serialize_field<T>(&mut self, key: &'static str, value: &T) -> Result<(), InvalidValue>
     where
-        T: Serialize,
+        T: Serialize + ?Sized,
     {
-        self.map.insert(KeyRef::Str(key), transform(value));
+        self.map.insert(key.into(), transform(value));
         Ok(())
     }
 
     fn end(self) -> Result<Value, InvalidValue> {
         let mut rv = BTreeMap::new();
-        rv.insert(
-            self.variant,
-            Value::from(ValueRepr::Map(Arc::new(self.map), MapType::Normal)),
-        );
+        rv.insert(self.variant.to_string(), Value::from_object(self.map));
         Ok(rv.into())
     }
 }
